/* xenv - expand environment variables in input files  -*- c -*-
   Copyright (C) 2021-2022 Sergey Poznyakoff

   Xenv is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Xenv is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with xenv. If not, see <http://www.gnu.org/licenses/>. */
%{
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <sysexits.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <poll.h>

char const *progname;       /* Program name */
dev_t device;               /* Device number of the filename. */
ino_t inode;                /* Inode number of the filename. */
int undef_error_option;     /* Treat undefined variables as error. */
int synclines_option;       /* Generate `#line NUM "FILE"' lines. */
int retain_unexpanded_option;/* Retain unexpanded constructs in the output. */
unsigned command_timeout = 0;  /* Timeout for command substitution. */
int status = 0;             /* Exit status */
int sigil = '$';            /* Current sigil. */
char const sigil_chars[] = "$@%&#";
			    /* Characters that are allowed for use as a sigil.
			       Note, that these are duplicated in the SPECIAL
			       character class definition below. */
char *env_prefix;           /* -e PREFIX */
static int env_prefix_len;  /* Number of prefix characters collected this far */

static int bracecount = 0;  /* Curly brace nesting level. */

struct point {
	char const *file;
	int line;
	int column;
};

struct locus {
	struct point beg, end;
};

static void push_state_at_point(int newstate, struct point const *point);
static void push_state(int newstate);
static void pop_state(void);
static struct point const *start_point(void);

static char *findenv(char const *ident, int len);
static void expandenv(char const *ident, int len);

enum {  /* Variable expansion modifiers. */
	EXP_DEFVAL = '-',   /* Use default value. */
	EXP_ASSIGN = '=',   /* Assign default value. */
	EXP_ERROR  = '?',   /* Display error if null or unset. */
	EXP_ALTER  = '+',   /* Use alternate value. */
	EXP_TERNARY = '|',  /* Ternary operator. */
	EXP_SET = 256       /* Special case: the $$set statement */
};

static void expand_inline_push(int type, char *varname, int varlen, int suppress);
static int expand_inline_pop(void);
static void expand_inline_flop(void);

static int cond_peek(struct point *);
static void cond_push(int result);
static int cond_pop(void);

static void push_input(char const *name, int oknotfound, struct locus *loc);

static int parse_line_directive(char *text);

static void syncline(int);
static void syncline_flush(void);
static void echo(char const*, size_t);
#define ECHO echo(yytext, yyleng)

struct filename {
	struct filename *next;
	size_t len;
	char name[1];
};

static struct filename *filename_head;

static void *xmalloc(size_t size);

static char const *
add_filename(char const *name, size_t len)
{
	struct filename *f;
	for (f = filename_head; f; f = f->next)
		if (f->len == len && memcmp(f->name, name, len) == 0)
			return f->name;
	f = xmalloc(sizeof(*f) + len);
	memcpy(f->name, name, len);
	f->name[len] = 0;
	f->len = len;
	f->next = filename_head;
	filename_head = f;
	return f->name;
}

static void
point_init(struct point *point, char const *name, size_t len)
{
	point->file = add_filename(name, len);
	point->line = 1;
	point->column = 0;
}

static void
point_advance_line(struct point *point, int n)
{
	point->line += n;
	point->column = 0;
}

static struct point curpoint;
static struct locus curlocus;

#define YY_USER_ACTION do {						\
		curpoint.column += yyleng;				\
		curlocus.end = curpoint;				\
	} while (0);

static void
point_print(FILE *fp, struct point const *pt)
{
	if (pt->column == 0)				
		fprintf(fp, "%s:%u",				
			pt->file,				
			pt->line);			
	else							
		fprintf(fp, "%s:%u.%u",			
			pt->file,				
			pt->line,				
			pt->column);			
}

static void
locus_print(FILE *fp, struct locus const *loc)
{
	if (loc->beg.column == 0)				
		fprintf(fp, "%s:%u",				
			loc->beg.file,				
			loc->beg.line);			
	else if (strcmp(loc->beg.file, loc->end.file))	
		fprintf(fp, "%s:%u.%u-%s:%u.%u",		
			loc->beg.file,				
			loc->beg.line, loc->beg.column,	
			loc->end.file,				
			loc->end.line, loc->end.column);	
	else if (loc->beg.line != loc->end.line)		
		fprintf(fp, "%s:%u.%u-%u.%u",			
			loc->beg.file,				
			loc->beg.line, loc->beg.column,	
			loc->end.line, loc->end.column);	
	else if (loc->beg.column != loc->end.column)		
		fprintf(fp, "%s:%u.%u-%u",			
			loc->beg.file,				
			loc->beg.line, loc->beg.column,	
			loc->end.column);			
	else							
		fprintf(fp, "%s:%u.%u",			
			loc->beg.file,				
			loc->beg.line,				
			loc->beg.column);			
}

static void
error_at_point(struct point const *pt, char const *fmt, ...)
{
	va_list ap;
        point_print(stderr, pt);
	fprintf(stderr, ": ");
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fputc('\n', stderr);
}

static void
error_at_locus(struct locus const *loc, char const *fmt, ...)
{
	va_list ap;
	if (!loc)
		loc = &curlocus;
	locus_print(stderr, loc);
	fprintf(stderr, ": ");
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fputc('\n', stderr);
}

static void
xenv_error(int status, int errnum, char const *fmt, ...)
{
	va_list ap;
	fprintf(stderr, "%s: ", progname);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	if (errnum)
		fprintf(stderr, ": %s", strerror(errnum));
	fputc('\n', stderr);
	if (status)
		exit(status);
}

/*
 * TEXT is a portion of preprocessor statement after double sigil marker.
 * Find the variable identifier part.  Return a pointer to the identifier
 * and store its length in LEN.
 * If KW is supplied, initialize it with the pointer to the preprocessor
 * keyword (after $$ and optional whitespace).
 */
static char *
find_ident(char *text, int *len, char **kw, struct locus *loc)
{
	char *start = text;
	if (text[0] == ' ' || text[0] == '\t')
		text += strspn(text, " \t");
	if (kw)
		*kw = text;
	text += strcspn(text, " \t");
	text += strspn(text, " \t");
	*len = strcspn(text, " \t\n");
	if (loc) {
		loc->beg = curlocus.beg;
		loc->beg.column += 2 + text - start;
		loc->end = loc->beg;
		loc->end.column += *len - 1;
	}
	return text;
}

/*
 * Memory management with error handling.
 */
static void
enomem(void)
{
	xenv_error(0, 0, "not enough memory");
	abort();
}

static void *
xmalloc(size_t size)
{
	char *p = malloc(size);
	if (!p)
		enomem();
	return p;
}

static void *
xcalloc(size_t nmemb, size_t size)
{
	char *p = calloc(nmemb, size);
	if (!p)
		enomem();
	return p;
}

static char *
xrealloc(char *ptr, size_t size)
{
	char *p = realloc(ptr, size);
	if (!p)
		enomem();
	return p;
}

static void *
x2nrealloc(void *p, size_t *pn, size_t s)
{
	size_t n = *pn;
	char *newp;

	if (!p) {
		if (!n) {
			/* The approximate size to use for initial small
			   allocation requests, when the invoking code
			   specifies an old size of zero.  64 bytes is
			   the largest "small" request for the
			   GNU C library malloc.  */
			enum { DEFAULT_MXFAST = 64 };

			n = DEFAULT_MXFAST / s;
			n += !n;
		}
	} else if ((size_t) -1 / 3 * 2 / s <= n)
		enomem();
	else
		n += (n + 1) / 2;

	newp = realloc(p, n * s);
	if (!newp)
		enomem();
	*pn = n;
	return newp;
}


struct xenv_feature {
	char *name;
	int enabled;
};

enum {
	FEATURE_COMMAND,
	FEATURE_COMMENT,
	FEATURE_DIRECTIVE,
	FEATURE_QUOTE,
	FEATURE_SIGIL
};

static struct xenv_feature feature[] = {
	{ "command",     1 },
	{ "comment",     1 },
	{ "directive",   1 },
	{ "quote",       1 },
	{ "sigil",       -1 },
	{ NULL },
};

static int
feature_set(char const *name)
{
	int i, val = 1;

	if (strncmp(name, "no-", 3) == 0) {
		name += 3;
		val = 0;
	}
	if (*name) {
		for (i = 0; feature[i].name; i++) {
			if (strcmp(feature[i].name, name) == 0) {
				feature[i].enabled = val;
				return 0;
			}
		}
	}
	return -1;
}

static void expandcom(char *str, struct locus const *locus);

static int collect_delimited_text(int od, int cd,
				  void (*collector)(void *, int),
				  void *closure);

static void
collect_echo(void *closure, int c)
{
	fputc(c, yyout);
}

struct textacc {
	char *text;
	size_t len;
	size_t size;
};

static void
collect_text(void *closure, int c)
{
	struct textacc *ap = closure;
	if (ap->len == ap->size)
		ap->text = x2nrealloc(ap->text, &ap->size,
				      sizeof(ap->text[0]));
	ap->text[ap->len++] = c;
}

/*
 * General-purpose stack.
 */
typedef struct xenv_stack {
	size_t entry_size;
	char *stack;
	size_t max_stack;
	size_t tos;
} XENV_STACK;

#define XENV_STACK_INITIALIZER(type)  { sizeof(type), NULL, 0, 0 }
#define XENV_INIT_SIZE 64

static inline int
xenv_stack_size(XENV_STACK *stk)
{
	return stk->tos;
}

static void *
xenv_stack_peek(XENV_STACK *stk, int dec)
{
	if (-dec > stk->tos)
		return NULL;
	return stk->stack + stk->entry_size * (stk->tos + dec);
}

static void *
xenv_stack_push(XENV_STACK *stk)
{
	if (stk->tos == stk->max_stack) {
		if (!stk->stack)
			stk->stack = xmalloc(stk->max_stack = XENV_INIT_SIZE);
		else
			stk->stack = x2nrealloc(stk->stack, &stk->max_stack,
						sizeof(stk->stack[0]));
	}
	++stk->tos;
	return xenv_stack_peek(stk, -1);
}

static void *
xenv_stack_pop(XENV_STACK *stk)
{
	if (stk->tos == 0) {
		error_at_point(&curpoint, "out of popup space");
		exit(EX_SOFTWARE);
	}
	--stk->tos;
	return xenv_stack_peek(stk, 0);
}

%}
  /* Identifier */
IDENT [a-zA-Z_][a-zA-Z_0-9]*
  /* Whitespace */
WS [ \t][ \t]*
  /* Optional whitespace */
OWS [ \t]*
  /* Special character: can be a sigil */
SPECIAL [$@%&#]

  /*
   * States
   */
%x COMMENT FALSE SIGIL DSIGIL VERBATIM SETEXP SETVAL ALT DQ SQ PREFIX
  /*
   * Possible state transitions are:
   *
   *  State     Input          New state         Comment
   *  --------  ------------   ---------------   -------
   *   INITIAL  <sigil>        SIGIL             Unless -e is used
   *   INITIAL  <sigil>        PREFIX            If -e is used
   *   INITIAL  <two sigils>   DSIGIL            Only at the beginning of
   *                                             line, optionally preceded
   *                                             by whitespace.
   *                                             Disabled by -Wno-directive.
   *   PREFIX   env_prefix     SIGIL
   *   SIGIL    <sigil>{<ID>   ALT               ID stands for identifier
   *                                             followed by optional ?
   *                                             and one of -, =, ?, +, |
   *   SIGIL    {*             COMMENT           Disabled by -Wno-comment
   *   ALT      '"'            DQ
   *            "'"            SQ
   *   DQ       <sigil>        SIGIL
   *   COMMENT  *}             INITIAL
   *   DSIGIL   "verbatim"     VERBATIM          Unless prior state was FALSE.
   *   DSIGIL   "verbatim"     FALSE             If prior state was FALSE.
   *   DSIGIL   "set IDENT$"   INITIAL
   *            'set IDENT "'  SETEXP
   *            "set IDENT '"  SETVAL
   *            "ifset"        INITIAL or FALSE  Depending on condition.
   *            "ifnset"                         --"--
   *            "ifdef"                          --"--
   *            "ifndef"                         --"--
   *            "else"                           --"--
   *   VERBATIM "end"          INITIAL
   *   FALSE    "else"         INITIAL
   *            "endif"
   *   SETEXP   '"'            INITIAL           pop state
   *   SETEXP   <sigil>        SIGIL
   *   SETVAL   "'"            INITIAL           pop state
   */
%%
<INITIAL,ALT,DQ,SETEXP,SETVAL>{
\n+  {
    point_advance_line(&curpoint, yyleng);
    ECHO;
 }
}

<ALT,DQ,SETEXP>{
\{  { bracecount++; ECHO; }
\}  { if (expand_inline_pop())
	      ECHO;
      else if (YYSTATE == ALT)
	      pop_state();
      bracecount--;
 }
\|  { expand_inline_flop(); }
}

<INITIAL,ALT,DQ,SETEXP>{
{SPECIAL} {
	   if (yytext[0] == sigil) {
		curlocus.beg = curpoint;
		if (env_prefix){
			env_prefix_len = 0;
			push_state(PREFIX);
		} else {
			push_state(SIGIL);
		}
	   } else
		ECHO;
	 }

}

<INITIAL>{
\\. {
	if (yytext[1] == '\\' || yytext[1] == sigil)
		fputc(yytext[1], yyout);
	else
		ECHO;
 }
^"#line ".*\n {
	if (parse_line_directive(yytext)) {
		point_advance_line(&curpoint, 1);
	}
	syncline(0);
	ECHO;
 }

^{OWS}{SPECIAL}{2} {
	if (feature[FEATURE_DIRECTIVE].enabled &&
	    yytext[yyleng-2] == sigil) {
		curlocus.beg = curpoint;
		if (yytext[yyleng-1] == sigil) {
			/* Column points to the second '$'; adjust it. */
			curlocus.beg.column--;
			push_state(DSIGIL);
		} else {
			yyless(yyleng-1);
			echo(yytext, yyleng-2);
			push_state(SIGIL);
		}
	} else {
		ECHO;
	}
 }
}

<ALT>{
\\. {
	if (yytext[1] == '\\' || yytext[1] == sigil ||
	    yytext[1] == '\'' || yytext[1] == '"')
		fputc(yytext[1], yyout);
	else
		ECHO;
 }
\"     push_state(DQ);
"'"    push_state(SQ);
}

<DQ>{
\\. {
	if (yytext[1] == '\\' || yytext[1] == sigil ||
	    yytext[1] == '\'' || yytext[1] == '"')
		fputc(yytext[1], yyout);
	else
		ECHO;
 }
\"      pop_state();
}

<SQ>{
\'      pop_state();
}

<COMMENT>{
[^*\n]*        /* eat anything that's not a '*' */
"*"+[^*}\n]*   /* eat up '*'s not followed by '}'s */
\n             point_advance_line(&curpoint, 1); /* Keep track of line numbers. */
"*"+"}"        { pop_state(); syncline(0); }
}

<PREFIX>{	
\n|. { if (yytext[0] == env_prefix[env_prefix_len]) {
	       if (env_prefix[++env_prefix_len] == 0) {
		       pop_state();
		       push_state(SIGIL);
	       }
       } else {
	       pop_state();
	       fputc(sigil, yyout);
	       echo(env_prefix, env_prefix_len);
	       ECHO;
	       if (yytext[0] == '\n')
		       point_advance_line(&curpoint, 1);
       }
  }
}

<SIGIL>{
"{*" { pop_state();
       if (feature[FEATURE_COMMENT].enabled) {
	       push_state_at_point(COMMENT, &curlocus.beg);
       } else {
	       fputc(sigil, yyout);
	       ECHO;
       }
 }
"[" {
	if (feature[FEATURE_QUOTE].enabled) {
		/*
		 * Save start_point for eventual use in the diagnostic
		 * message below
		 */
		struct point pt = *start_point();
		int c = collect_delimited_text('[', ']', collect_echo, yyout);
		pop_state();
		if (c == EOF) {
			error_at_point(&pt,
				       "end of file in inline verbatim text");
			return EOF;
		}
	} else {
		pop_state();
		fputc(sigil, yyout);
		ECHO;
	}
 }
"(" {
	if (feature[FEATURE_COMMAND].enabled) {
		struct textacc t = { NULL, 0, 0 };
		struct point pt = *start_point();
		int c = collect_delimited_text('(', ')', collect_text, &t);
		pop_state();
		collect_text(&t, 0);
		if (c == EOF) {
			error_at_point(&pt,
				       "end of file in command expansion");
			free(t.text);
			return EOF;
		}
		expandcom(t.text, &curlocus);
		free(t.text);
	} else {
		pop_state();
		fputc(sigil, yyout);
		ECHO;
	}
 }
{IDENT} { pop_state();
	  if (env_prefix) {
		  fputc(sigil, yyout);
		  echo(env_prefix, strlen(env_prefix));
		  ECHO;
	  } else {
		  expandenv(yytext, yyleng);
	  }
  }
\{{IDENT}\} { pop_state(); expandenv(yytext + 1, yyleng - 2); }
\{{IDENT}:?[-=?+|] {
	int len;
	char *val;
	int test_null = 0;
	int type = yytext[yyleng-1];
	int suppress;

	pop_state();
	if (yytext[yyleng-2] == ':') {
		test_null = 1;
		len = yyleng - 3;
	} else
		len = yyleng - 2;
	val = findenv(yytext + 1, len);
	if (val && !(test_null && *val == 0)) {
		if (type == EXP_TERNARY || type == EXP_ALTER) {
			suppress = 0;
		} else {
			fwrite(val, strlen(val), 1, yyout);
			suppress = 1;
		}
	} else if (type == EXP_TERNARY || type == EXP_ALTER) {
		suppress = 1;
	} else {
		suppress = 0;
		if (type == EXP_ERROR ) {
			int c = input();
			if (c == '}') {
				/* Provide default error message */
				if (test_null)
					error_at_locus(&curlocus,
						       "variable %.*s null or not set",
						       len, yytext + 1);
				else
					error_at_locus(&curlocus,
						       "variable %.*s not set",
						       len, yytext + 1);
				YY_BREAK;
			 }
			 unput(c);
		}
	}
	bracecount++;
	expand_inline_push(type, yytext + 1, len, suppress);
	push_state(ALT);
}

\n|.   {
	char s[2];
	s[0] = sigil;
	s[1] = yytext[0];
	echo(s, 2);
	pop_state();
	if (yytext[0] == '\n')
		point_advance_line(&curpoint, 1);
 }
}

<DSIGIL>{
{OWS}"verbatim"{OWS}"\n" {
	pop_state();
	if (YYSTATE != FALSE)
		push_state_at_point(VERBATIM, &curlocus.beg);
	point_advance_line(&curpoint, 1);
 }

{OWS}"sigil"{WS}{SPECIAL}{OWS}"\n" {
	struct locus loc = curlocus;
	loc.end.column--;

	switch (feature[FEATURE_SIGIL].enabled) {
	case 0:
		error_at_locus(&loc, "unrecognized directive (use of $$sigil disabled in the command line)");
		YY_BREAK;
		
	case -1:
		error_at_locus(&loc, "warning: use of the $$sigil directive is deprecated; consider using -e instead");
	}
	
	pop_state();
	if (YYSTATE != FALSE) {
		int len;
		char *ident = find_ident(yytext, &len, NULL, NULL);
		sigil = *ident;
	}
	point_advance_line(&curpoint, 1);
 }

{OWS}"unset"{WS}{IDENT}{OWS}\n {
	int len;
	char *ident = find_ident(yytext, &len, NULL, NULL);

	ident[len] = 0;
	unsetenv(ident);

	point_advance_line(&curpoint, 1);
	syncline(0);
	pop_state();
 }

{OWS}"set"{WS}{IDENT}{OWS}\n {
	int len;
	char *ident = find_ident(yytext, &len, NULL, NULL);

	ident[len] = 0;
	setenv(ident, "", 1);

	point_advance_line(&curpoint, 1);
	syncline(0);
	pop_state();
 }

{OWS}"set"{WS}{IDENT}{WS}\' {
	int len;
	char *ident = find_ident(yytext, &len, NULL, NULL);
	pop_state();
	push_state(SETVAL);
	expand_inline_push(EXP_SET, ident, len, 0);
 }

{OWS}"set"{WS}{IDENT}{WS}\" {
	int len;
	char *ident = find_ident(yytext, &len, NULL, NULL);
	pop_state();
	push_state(SETEXP);
	expand_inline_push(EXP_SET, ident, len, 0);
 }

{OWS}"if"n?(def|set){WS}{IDENT}{OWS}\n {
	int result;

	pop_state();
	result = cond_peek(NULL);
	if (result) {
		int len;
		char *ident;
		char *val;
		char *kw;
		int neg;

		ident = find_ident(yytext, &len, &kw, NULL);
		neg = kw[2] == 'n';
		val = findenv(ident, len);
		result = val != NULL && (kw[2+neg] == 'd' || *val != 0);
		if (neg)
			result = !result;
	}
	cond_push(result);
	point_advance_line(&curpoint, 1);
 }

{OWS}"else"{OWS}\n {
	pop_state();
	if (cond_peek(NULL) == -1) {
		error_at_locus(&curlocus, "%c%celse without matching %c%cif",
			       sigil, sigil, sigil, sigil);
		fputc(sigil, yyout);
		fputc(sigil, yyout);
		ECHO;
	} else {
		int result = cond_pop();
		if (cond_peek(NULL))
			result = !result;
		cond_push(result);
	}
	point_advance_line(&curpoint, 1);
}

{OWS}"endif"{OWS}\n {
	pop_state();
	if (cond_peek(NULL) == -1) {
		error_at_locus(&curlocus, "%c%cendif without matching %c%cif",
			       sigil, sigil, sigil, sigil);
		fputc(sigil, yyout);
		fputc(sigil, yyout);
		ECHO;
	} else {
		cond_pop();
	}
	point_advance_line(&curpoint, 1);
	syncline(0);
 }

{OWS}"source".*\n |
{OWS}"include".*\n {
	char *filename;
	int namelen;
	struct locus loc;
	
	pop_state();
	filename = find_ident(yytext, &namelen, NULL, &loc);
	filename[namelen] = 0;
	point_advance_line(&curpoint, 1);
	push_input(filename, 0, &loc);
	syncline(0);
 }

{OWS}"sinclude".*\n {
	char *filename;
	int namelen;
	struct locus loc;
	
	pop_state();
	filename = find_ident(yytext, &namelen, NULL, &loc);
	filename[namelen] = 0;
	push_input(filename, 1, &loc);
	syncline(0);
	point_advance_line(&curpoint, 1);
 }

. {
	pop_state();
	if (YYSTATE != FALSE) {
		char s[3];
		s[0] = s[1] = sigil;
		s[2] = yytext[0];
		error_at_locus(&curlocus, "unrecognized directive");
		echo(s, 3);
	}
 }
}

<SETEXP>{
\\. {
	if (yytext[1] == '\\' || yytext[1] == '"' || yytext[1] == sigil)
		fputc(yytext[1], yyout);
	else
		ECHO;
 }
\"{OWS}\n {
	syncline(1);
	expand_inline_pop();
	pop_state();
	point_advance_line(&curpoint, 1);
 }
\"/[^\n]+\n {
        error_at_locus(&curlocus, "extra characters after closing quote");
	expand_inline_pop();
	pop_state();
 }
}

<SETVAL>{
\'{OWS}\n {
	syncline(0);
	expand_inline_pop();
	pop_state();
	point_advance_line(&curpoint, 1);
 }
\'/[^\n]+\n {
        error_at_locus(&curlocus, "extra characters after closing quote");
	expand_inline_pop();
	pop_state();
 }
}


<FALSE>{
\n+    	point_advance_line(&curpoint, yyleng);
^{OWS}{SPECIAL}{2} {
	if (yytext[yyleng-2] == sigil && yytext[yyleng-1] == sigil) {
		curlocus.beg = curpoint;
		curlocus.beg.column--;
		push_state(DSIGIL);
	}
 }
. ;
}

<VERBATIM>{
\n+    { ECHO; point_advance_line(&curpoint, yyleng); }
^{OWS}{SPECIAL}{2}{OWS}"end"{OWS}"\n" {
	if (yytext[0] == sigil && yytext[1] == sigil) {
		pop_state();
		point_advance_line(&curpoint, 1);
	} else {
		ECHO;
	}
 }
}
%%
static struct point syncline_point;
static int beginning_of_line = 1;

static void
syncline(int off)
{
	if (synclines_option) {
		syncline_point = curpoint;
		if (off)
			point_advance_line(&syncline_point, off);
	}
}

static void
syncline_flush(void)
{
	if (beginning_of_line && syncline_point.line > 0 &&
	    (YYSTATE == INITIAL || YYSTATE == ALT)) {
		fprintf(yyout, "#line %d \"%s\"\n",
			syncline_point.line, syncline_point.file);
		syncline_point.line = 0;
	}
}

static void
echo(char const *text, size_t len)
{
	if (len == 0) return;
	syncline_flush();
	fwrite(text, len, 1, yyout);
	beginning_of_line = text[len-1] == '\n';
}

extern char **environ;

/*
 * Same as getenv(3), but uses first LEN bytes of IDENT as the name of
 * a variable.
 */
static char *
findenv(char const *ident, int len)
{
	size_t i;
	for (i = 0; environ[i]; i++) {
		size_t j;
		for (j = 0; j < len; j++)
			if (environ[i][j] != ident[j])
				break;
		if (j == len && environ[i][j] == '=')
			return environ[i] + j + 1;
	}
	return NULL;
}

/*
 * Expand environment variable whose name is given by first LEN bytes of
 * IDENT.
 */
static void
expandenv(char const *ident, int len)
{
	char *val = findenv(ident, len);

	syncline_flush();

	if (val) {
		echo(val, strlen(val));
		beginning_of_line = 0;
		if (strchr(val, '\n'))
			syncline(1);
	} else {
		if (undef_error_option) {
                        error_at_locus(&curlocus, "variable %.*s not defined",
				       len, ident);
			status = EX_DATAERR;
		}
		if (retain_unexpanded_option) {
			char c = sigil;
			echo(&c, 1);
			if (env_prefix)
				echo(env_prefix, strlen(env_prefix));
			ECHO;
		}
	}
}

struct saved_state {
	int state;
	struct point point;
};

static XENV_STACK state_stack = XENV_STACK_INITIALIZER(struct saved_state);

static struct point const *
start_point(void)
{
	struct saved_state *st = xenv_stack_peek(&state_stack, -1);
	return st ? &st->point : &curpoint;
}

static void
push_state_at_point(int newstate, struct point const *point)
{
	struct saved_state *st = xenv_stack_push(&state_stack);
	st->state = YYSTATE;
	st->point = *point;
	BEGIN(newstate);
}

static void
push_state(int newstate)
{
	push_state_at_point(newstate, &curpoint);
}

static void
pop_state(void)
{
	struct saved_state *st = xenv_stack_pop(&state_stack);
	BEGIN(st->state);
}

struct expand_inline {
	int type;      /* Expansion type (one of the EXP_ constants above). */
	int suppress;  /* Suppress output. */
	int bracecount;/* Brace nesting level this expansion was created at. */
	int bol;       /* Beginning of line status */
	FILE *fp;      /* Previous value of yyout */
	union {
		char *varname; /* type == EXP_ASSIGN - assign the expansion
				  to that variable */
		struct locus locus; /* type == EXP_ERROR - location of the error */
	};
};

static XENV_STACK expand_inline_stack = XENV_STACK_INITIALIZER(struct expand_inline);

/*
 * Push on stack the inline expansion entry of given TYPE.
 * VARNAME and VARLEN identify the variable if type is EXP_ASSIGN.
 * SUPPRESS is 1 if the output is to be suppressed and 0 otherwise.
 */
static void
expand_inline_push(int type, char *varname, int varlen, int suppress)
{
	struct expand_inline *ep;

	ep = xenv_stack_push(&expand_inline_stack);
	ep->type = type;
	switch (type) {
	case EXP_ASSIGN:
	case EXP_SET:
		assert(varname && varlen);
		ep->varname = xmalloc(varlen + 1);
		memcpy(ep->varname, varname, varlen);
		ep->varname[varlen] = 0;
		break;

	case EXP_ERROR:
		ep->locus = curlocus;
		break;
	}
	ep->fp = yyout;
	ep->suppress = suppress;
	ep->bracecount = bracecount;
	ep->bol = beginning_of_line;
	if (suppress) {
		yyout = fopen("/dev/null", "w+b");
		if (!yyout) {
			xenv_error(EX_OSERR, errno, "can't open /dev/null");
		}
	} else if ((yyout = tmpfile()) == NULL) {
		xenv_error(EX_OSERR, errno, "can't create temporary file");
	}
}

/*
 * Flop the suppression state of the ternary operator on the top of stack.
 * If there's no ternary on tos or its brace count doesn't match the current
 * one, echo yytext.
 */
static void
expand_inline_flop(void)
{
	struct expand_inline *ep = xenv_stack_peek(&expand_inline_stack, -1);
	if (ep &&
	    ep->bracecount == bracecount &&
	    ep->type == EXP_TERNARY) {
		int suppress = ep->suppress;
		expand_inline_pop();
		expand_inline_push(EXP_TERNARY, NULL, 0, !suppress);
	} else
		ECHO;
}
/*
 * Pop the topmost expansion entry off the stack.
 * Return 1 if the operation is not applicable (wrong state or brace level
 * mismatch).
 * Return 0 on success.
 */
static int
expand_inline_pop(void)
{
	struct expand_inline *ep;

	ep = xenv_stack_peek(&expand_inline_stack, -1);
	if (ep == NULL || ep->bracecount != bracecount) {
		return 1;
	}
	xenv_stack_pop(&expand_inline_stack);
	if (!ep->suppress) {
		if (ep->type == EXP_ASSIGN || ep->type == EXP_SET) {
			long size;
			char *value;
			size_t n;

			fseek(yyout, 0, SEEK_END);
			size = ftell(yyout);
			value = xmalloc(size + 1);
			rewind(yyout);
			n = fread(value, 1, size, yyout);
			if (n == -1) {
				xenv_error(EX_OSERR, errno, "read error");
			} else if (n > 0 && n < size) {
				xenv_error(EX_OSERR, 0, "short read");
			}
			value[n] = 0;
			setenv(ep->varname, value, 1);
			if (ep->type == EXP_ASSIGN)
				fwrite(value, size, 1, ep->fp);
			if (strchr(value, '\n'))
				syncline(1);
		} else if (ep->type == EXP_ERROR) {
			int c;

			ep->locus.end = curpoint;
			locus_print(stderr, &ep->locus);
			fprintf(stderr, ": ");
			rewind(yyout);
			while ((c = fgetc(yyout)) != EOF)
				fputc(c, stderr);
			fputc('\n', stderr);
		} else {
			int c;

			rewind(yyout);
			while ((c = fgetc(yyout)) != EOF) {
				fputc(c, ep->fp);
				if (c == '\n')
					syncline(1);
			}
		}
	} else
		/* FIXME: Emit syncline only if there were newlines in the
		   suppressed text. */
		syncline(1);

	if (ep->type == EXP_ASSIGN || ep->type == EXP_SET)
		free(ep->varname);
	fclose(yyout);
	yyout = ep->fp;
	beginning_of_line = ep->bol;
	return 0;
}

/*
 * Condition stack.  Implements conditional preprocessor blocks like:
 *
 *   $$ifset
 *    TEXT1
 *   $$else
 *    TEXT2
 *   $$endif
 */

struct cond_state {
	int result;
	struct point point;
};

static XENV_STACK cond_stack = XENV_STACK_INITIALIZER(struct cond_state);

/*
 * Return the evaluation result from the top of stack.
 */
static int
cond_peek(struct point *point)
{
	struct cond_state *st = xenv_stack_peek(&cond_stack, -1);
	if (st == NULL)
		return -1;
	if (point)
		*point = st->point;
	return st->result;
}

/*
 * Push RESULT on stack. Switch to state FALSE if result is false. Otherwise,
 * output a syncline.
 */
static void
cond_push(int result)
{
	struct cond_state *st = xenv_stack_push(&cond_stack);
	st->result = result;
	st->point = curlocus.beg;
	if (!result)
		push_state_at_point(FALSE, &curlocus.beg);
	else
		syncline(1);
}

/*
 * Pop the entry off the stack.  If it is false, restore previous scanner
 * state.  Return the evaluation result.
 */
static int
cond_pop(void)
{
	struct cond_state *st = xenv_stack_pop(&cond_stack);
	if (!st->result)
		pop_state();
	return st->result;
}

static int
parse_line_directive(char *text)
{
	unsigned long n;
	char *fname, *p;

	text += sizeof("#line ")-1;

	n = strtoul(text, &p, 10);
	if ((n == 0 && errno == ERANGE) || n > INT_MAX ||
	    !(*p == 0 || *p == ' ' || *p == '\t' || *p == '\n'))
		return 1;
	while (*p == ' ' || *p == '\t')
		p++;
	if (*p == '\n' || *p == 0) {
		curpoint.line = n;
		curpoint.column = 0;
		return 0;
	}
	if (*p != '"')
		return 1;
	p++;
	fname = p;
	p = strchr(fname, '"');
	if (!p)
		return 1;
        point_init(&curpoint, fname, p - fname);
	curpoint.line = n;
	curpoint.column = 0;
	curlocus.beg = curlocus.end = curpoint;
	return 0;
}

/*
 * Input file handling.
 */
static char **input_files; /* Array of input files. */
static int input_index;    /* Index of the next file in input_files. */

static int
file_error_exit(void)
{
	switch (errno) {
	case ENOENT:
		exit(EX_NOINPUT);

	case EACCES:
		exit(EX_NOPERM);

	default:
		exit(EX_OSERR);
	}
}

static void
open_input(char const *name)
{
	FILE *fp;
	struct stat st;

	if (name[0] == '-' && name[1] == 0) {
		name = "<stdin>";
		fp = stdin;
	} else {
		if ((fp = fopen(name, "r")) == NULL) {
			xenv_error(0, errno, "can't open %s", name);
			file_error_exit();
		}
	}

	yyin = fp;
        point_init(&curpoint, name, strlen(name));
	beginning_of_line = 1;
	syncline(0);

	if (fstat(fileno(yyin), &st)) {
		xenv_error(0, errno, "can't stat file %s", name);
		file_error_exit();
	}
	device = st.st_dev;
	inode = st.st_ino;
}

struct search_path_dir {
	struct search_path_dir *next;
	size_t len;
	char name[1];
};

struct search_path_dir *search_path_head, *search_path_tail;

static void
search_path_append(char *dir)
{
	size_t namelen = strlen(dir);
	struct search_path_dir *dp = malloc(sizeof(*dp) + namelen);
	dp->len = namelen;
	strcpy(dp->name, dir);
	dp->next = NULL;
	if (search_path_tail)
		search_path_tail->next = dp;
	else
		search_path_head = dp;
	search_path_tail = dp;
}

static char *
find_file(char const *name)
{
	struct search_path_dir *dp;
	size_t namelen = strlen(name);

	char *filebuf = NULL;
	size_t filebuflen = 0;

	for (dp = search_path_head; dp; dp = dp->next) {
		size_t len = dp->len + namelen + 2;
		if (len > filebuflen) {
			filebuflen = len;
			filebuf = xrealloc(filebuf, filebuflen);
		}
		memcpy(filebuf, dp->name, dp->len);
		filebuf[dp->len] = '/';
		strcpy(filebuf + dp->len + 1, name);

		if (access(filebuf, F_OK) == 0)
			return filebuf;
		else if (errno != ENOENT)
			error_at_locus(&curlocus, "can't stat %s: %s",
				       filebuf, strerror(errno));
	}
	free(filebuf);
	return NULL;
}

struct context_stack {
	YY_BUFFER_STATE state;
	char *filename;
	struct locus locus;
	ino_t inode;
	dev_t device;
	struct context_stack *prev;
};

static struct context_stack *context_stack_top;

static struct context_stack *
context_find(dev_t device, ino_t inode)
{
	struct context_stack *ctx = context_stack_top;
	while (ctx) {
		if (ctx->inode == inode && ctx->device == device)
			break;
		ctx = ctx->prev;
	}
	return ctx;
}

static void
push_input(char const *name, int oknotfound, struct locus *loc)
{
	struct context_stack *sp;
	FILE *fp;
	char *nameptr = NULL;
	struct stat st;
	int found;

	if (!(name[0] == '/' ||
	      (name[0] == '.' && name[1] == '/') ||
	      (name[0] == '.' && name[1] == '.' && name[2] == '/'))) {
		nameptr = find_file(name);
		if (!nameptr) {
			if (oknotfound) {
				return;
			}
			error_at_locus(loc, "file not found: %s", name);
			exit(EX_NOINPUT);
		}
		name = nameptr;
	}

	if (stat(name, &st)) {
		if (oknotfound && errno == ENOENT) {
			free(nameptr);
			return;
		}

                error_at_locus(loc, "can't stat file %s: %s",
			       name, strerror(errno));
		file_error_exit();
	}

	if (st.st_ino == inode && st.st_dev == device) {
		found = 1;
		sp = context_stack_top;
	} else if ((sp = context_find(st.st_dev, st.st_ino)) != NULL) {
		sp = sp->prev;
		found = 1;
	} else
		found = 0;

	if (found) {
		error_at_locus(loc, "file %s already included", name);
		if (sp)
			error_at_locus(&sp->locus,
				       "place of the initial inclusion");
		exit(EX_UNAVAILABLE);
	}

	fp = fopen(name, "r");
	if (!fp) {
		if (oknotfound && errno == ENOENT) {
			free(nameptr);
			return;
		}
		error_at_locus(loc, "can't open file %s: %s",
			       name, strerror(errno));
		file_error_exit();
	}

	sp = xcalloc(1, sizeof(*sp));
	sp->state = YY_CURRENT_BUFFER;
	sp->locus = *loc;
	sp->device = device;
	sp->inode = inode;
	sp->prev = context_stack_top;
	context_stack_top = sp;

	yyin = fp;
	yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
	point_init(&curpoint, name, strlen(name));
	beginning_of_line = 1;
	device = st.st_dev;
	inode = st.st_ino;

	free(nameptr);
}

static int
pop_input(void)
{
	struct context_stack *sp;

	if (yyin != stdin)
		fclose(yyin);
	yy_delete_buffer(YY_CURRENT_BUFFER);

	sp = context_stack_top;
	if (!sp)
		return 1;
	context_stack_top = sp->prev;

	yy_switch_to_buffer(sp->state);
        curpoint = sp->locus.end;
	point_advance_line(&curpoint, 1);
	beginning_of_line = 1;
	syncline(0);

	device = sp->device;
	inode = sp->inode;

	free(sp->filename);
	free(sp);

	return 0;
}

int
yywrap(void)
{
	if (pop_input() == 0)
		return 0;
	if (input_files[input_index]) {
		open_input(input_files[input_index++]);
		return 0;
	}
	return 1;
}

/*
 * Process input from the current position up to and including the closing
 * delimiter (CD).  It is supposed that the function is called immediately
 * after the opening delimiter (OD) was seen.  For each consumed character,
 * call COLLECTOR with CLOSURE as the first argument and the character read
 * as the second.  Assume each OD is paired by CD.
 */
static int
collect_delimited_text(int od, int cd, void (*collector)(void *, int),
		       void *closure)
{
	int nesting = 1;
	int c;
	/*
	 * On end of file, input() returns EOF in flex versions
	 * before 2.6.1, and 0 in flex 2.6.1 and later.
	 * See https://github.com/westes/flex/issues/448
	 */
	while ((c = input()) != EOF && c != 0) {
		curpoint.column++;
		if (c == od)
			nesting++;
		else if (c == cd) {
			if (--nesting == 0)
				break;
		} else if (c == '\n')
			point_advance_line(&curpoint, 1);
		collector(closure, c);
	}
	curlocus.end = curpoint;
	return c;
}

/*
 * Replace each occurrence of the current sigil in STR with $.
 * Don't touch text between single quotes.
 */
static void
replsig(char *str)
{
	int inq = 0;
	while (*str) {
		if (inq) {
			if (*str == '\'')
				inq = 0;
		} else {
			if (*str == '\\')
				str++;
			else if (*str == '\'')
				inq = 1;
			else if (*str == sigil)
				*str = '$';
		}
		str++;
	}
}

static inline int
timeval_cmp(struct timeval const *a, struct timeval const *b)
{
	if (a->tv_sec < b->tv_sec)
		return -1;
	if (a->tv_sec > b->tv_sec)
		return 1;
	if (a->tv_usec < b->tv_usec)
		return -1;
	if (a->tv_usec > b->tv_usec)
		return 1;
	return 0;
}

static inline int
timeval_diff(struct timeval const *a, struct timeval const *b)
{
	struct timeval d;

	d.tv_sec = a->tv_sec - b->tv_sec;
	d.tv_usec = a->tv_usec - b->tv_usec;
	if (d.tv_usec < 0) {
		--d.tv_sec;
		d.tv_usec += 1000000;
	}
	return d.tv_sec * 1000 + d.tv_usec / 1000;
}

static int
countnl(char const *s)
{
	int n = 0;
	for (; *s; s++)
		if (*s == '\n')
			n++;
	return n;
}

/*
 * Run the command in STR, redirecting its standard output to yyout.
 * Take care to remove trailing newlines, however.
 * If command_timeout is set, make sure the command is terminated if
 * it runs longer than that.
 */
static void
expandcom(char *str, struct locus const *locus)
{
	pid_t pid;
	FILE *fperr;
	int p[2];

	syncline_flush();

	if (sigil != '$')
		replsig(str);

	fperr = tmpfile();
	if (!fperr) {
		xenv_error(EX_OSERR, errno,
			   "can't create temporary error output");
	}

	if (pipe(p)) {
		xenv_error(EX_OSERR, errno, "pipe");
	}

	pid = fork();
	if (pid == -1) {
		xenv_error(EX_OSERR, errno, "fork");
	}

	if (pid == 0) {
		/* Child */
		char *argv[4];
		argv[0] = getenv("SHELL");
		if (!argv[0])
			argv[0] = "/bin/sh";
		argv[1] = "-c";
		argv[2] = (char *)str;
		argv[3] = NULL;

		close(0);
		if (open("/dev/null", O_RDONLY) == -1) {
			xenv_error(EX_OSERR, errno, "can't open /dev/null");
		}

		if (dup2(p[1], 1) == -1) {
			xenv_error(EX_OSERR, errno, "dup stdout");
		}
		close(p[1]);

		if (dup2(fileno(fperr), 2) == -1) {
			xenv_error(EX_OSERR, errno, "dup stderr");
		}
		execvp(argv[0], argv);
		_exit(127);
	} else {
		/* master */
		struct pollfd pfd;
		struct timeval end, now;
		int ex;
		int nl = 0;
		int outnl = 0;
		int c;
		pid_t pn;
		int sigsent = 0;

		close(p[1]);

		gettimeofday(&end, NULL);
		end.tv_sec += command_timeout;

		pfd.fd = p[0];
		pfd.events = POLLIN;
		nl = 0;
		while (1) {
			int n, to;
			char ch;

			if (pid != -1) {
				pn = waitpid(pid, &ex, WNOHANG);
				if (pn == pid) {
					pid = -1;
				} else if (pn == -1) {
					xenv_error(EX_OSERR, errno, "waitpid");
				}
			}

			if (command_timeout) {
				gettimeofday(&now, NULL);
				if (timeval_cmp(&now, &end) >= 0) {
					error_at_locus(locus,
						       "command timed out");
					kill(pid, SIGKILL);
					sigsent = 1;
					break;
				}
				to = timeval_diff(&end, &now);
			} else
				to = 0;

			n = poll(&pfd, 1, to);
			if (n == -1) {
				if (errno == EINTR)
					continue;
				xenv_error(EX_OSERR, errno, "poll");
			}
			if (pfd.revents & POLLIN) {
				n = read(pfd.fd, &ch, 1);
				if (n == -1) {
					xenv_error(EX_OSERR, errno, "read");
				} else if (n == 0) {
					break;
				} else if (ch == '\n') {
					nl++;
				} else {
					if (nl) {
						outnl += nl;
						do
							fputc('\n', yyout);
						while (--nl > 0);
					}
					fputc(ch, yyout);
				}
			} else if (pfd.revents == POLLHUP)
				break;
		}
		close(pfd.fd);

		if (outnl != countnl(str))
			syncline(1);

		if (pid != -1) {
			pn = waitpid(pid, &ex, 0);
			if (pn == -1) {
				xenv_error(EX_OSERR, errno, "waitpid");
			}
		}

		if (WIFEXITED(ex) && WEXITSTATUS(ex))
			error_at_locus(locus, "command exited with status %d",
				       WEXITSTATUS(ex));
		else if (WIFSIGNALED(ex) && !(WTERMSIG(ex) == SIGKILL && sigsent)) {
			error_at_locus(locus,
				       "command terminated on signal %d",
				       WTERMSIG(ex));
		}

		/* Translate error output */
		rewind(fperr);
		nl = 1; /* Re-use nl as a beginning-of-line indicator */
		while ((c = fgetc(fperr)) != EOF) {
			if (nl) {
				locus_print(stderr, locus);
				fprintf(stderr, ": ");
				nl = 0;
			}
			fputc(c, stderr);
			nl = c == '\n';
		}
		fclose(fperr);
		beginning_of_line = 0;
	}
}

static void
usage(FILE *fp)
{
	int i;

	fprintf(fp,
		"usage: %s [-hmnsux] [-e NAME] [-D NAME[=VALUE]] [-p COMMAND] [-S CHAR]\n"
		"       [-U NAME] [-W [no-]FEATURE] FILE [FILE...]\n", progname);
	fprintf(fp, "expands environment variables and shell commands in input files\n");
	fprintf(fp, "\nOPTIONS are:\n\n");
	fprintf(fp, "  -D NAME[=VALUE]  set environment variable NAME\n");
	fprintf(fp, "  -I DIRECTORY     append DIRECTORY to include path\n");
	fprintf(fp, "  -m               pipe output to m4\n");
	fprintf(fp, "  -n               don't produce output, only report errors\n");
	fprintf(fp, "  -e ENV           define environment meta-variable;\n"
		    "                   variables are to be referenced as $ENV{NAME}\n");
	fprintf(fp, "  -p COMMAND       pipe output to COMMAND\n");
	fprintf(fp, "  -r               retain unexpanded constructs in the output\n");
	fprintf(fp, "  -S CHAR          change sigil character (allowed chars: %s);\n", sigil_chars);
	fprintf(fp, "                   this option is deprecated and will be removed in the future\n");
	fprintf(fp, "  -s               generate `#line NUM \"FILE\"' lines\n");
	fprintf(fp, "  -t SECONDS       timeout for command substitution\n");
	fprintf(fp, "  -U NAME          unset environment variable NAME\n");
	fprintf(fp, "  -u               treat unset variables as errors\n");
	fprintf(fp, "  -v               print the version number and exit\n");
	fprintf(fp, "  -W [no-]FEATURE  enable or disable specific xenv feature\n");
	fprintf(fp, "  -x               enable debugging\n");
	fprintf(fp, "  -h or -?         print this help text\n");
	fprintf(fp, "\n");
	fprintf(fp, "FEATURE names are:\n");
	for (i = 0; feature[i].name; i++)
		fprintf(fp, "  %s\n", feature[i].name);
	fprintf(fp, "\n");

	fprintf(fp, "Report bugs and suggestions to <gray@gnu.org>\n");
}

static int copyright_year = 2022;

static void
version(void)
{
	printf("%s (%s) %s\n", progname, PACKAGE_NAME, PACKAGE_VERSION);
	printf("Copyright (C) %d Sergey Poznyakoff\n", copyright_year);
	printf("\
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\nThis is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law.\n\
");
}

int
main(int argc, char **argv)
{
	int c;
	int dry_run = 0;
	char *pipe_command = NULL;
	int m4_option = 0;
	char *sigil_option = NULL;
	
	progname = argv[0];
	yyset_debug(0);
	while ((c = getopt(argc, argv, "D:e:hI:mnp:rS:st:U:uvW:x?")) != EOF) {
		switch (c) {
		case 'D':{
			char *p = strchr(optarg, '=');
			if (p)
				*p++ = 0;
			else
				p = "";
			setenv(optarg, p, 1);
			break;
		}

		case 'I':
			search_path_append(optarg);
			break;

		case 'U':
			unsetenv(optarg);
			break;

		case 'e':
			env_prefix = optarg;
			break;

		case 'h':
			usage(stdout);
			return 0;

		case 'm':
			m4_option = 1;
			break;

		case 'n':
			dry_run = 1;
			break;

		case 'p':
			m4_option = 0;
			pipe_command = optarg;
			break;

		case 'r':
			retain_unexpanded_option = 1;
			break;

		case 'S':
			sigil_option = optarg;
			break;

		case 's':
			synclines_option = 1;
			break;

		case 't': {
			unsigned long n;
			char *p;

			errno = 0;
			n = strtoul(optarg, &p, 10);
			if (errno || n >= UINT_MAX) {
				xenv_error(EX_USAGE, errno,
					   "invalid timeout value");
			}
			command_timeout = n;
			break;
		}

		case 'u':
			undef_error_option = 1;
			break;

		case 'v':
			version();
			return 0;

		case 'x':
			yyset_debug(1);
			break;

		case 'W':
			if (feature_set(optarg)) {
				xenv_error(EX_USAGE, 0,
					   "invalid feature name: %s", optarg);
			}
			break;

		default:
			if (optopt == '?') {
				usage(stdout);
				return 0;
			}
			usage(stderr);
			return EX_USAGE;
		}
	}

	if (sigil_option) {
		switch (feature[FEATURE_SIGIL].enabled) {
		case 0:
			xenv_error(EX_USAGE, 0,
				   "the use of the -S option is disabled; use -e instead (or use -Wsigil to enable it)");
			break;

		case -1:
			xenv_error(0, 0,
				   "warning: the use of the -S option is deprecated; consider using -e instead");
		}
		if (sigil_option[1]) {
			xenv_error(EX_USAGE, 0, "sigil too long");
		}
		if (!strchr(sigil_chars, sigil_option[0])) {
			xenv_error(EX_USAGE, 0,
				   "%s is not allowed as a sigil; allowed characters are: %s",
				   sigil_option, sigil_chars);
		}
		sigil = sigil_option[0];
	}
	
	if (m4_option) {
		if (synclines_option)
			pipe_command = "m4 -s";
		else
			pipe_command = "m4";
	}

	if (dry_run) {
		int fd;

		if ((fd = open("/dev/null", O_WRONLY)) == -1 ||
		    dup2(fd, 1) == -1) {
			xenv_error(EX_OSERR, errno,
				   "can't open /dev/null for writing");
			return EX_OSERR;
		}
		close(fd);
	}

	input_files = argv + optind;
	input_index = 0;
	if (input_files[0] == NULL) {
		static char *stdin_file[] = { "-", NULL };
		input_files = stdin_file;
	}

	if (pipe_command) {
		yyout = popen(pipe_command, "w");
		if (!yyout) {
			xenv_error(EX_OSERR, errno, "popen");
		}

	} else {
		yyout = stdout;
	}
	open_input(input_files[input_index++]);
	yylex();

	switch (YYSTATE) {
	case INITIAL: {
		struct point pt;
		if (cond_peek(&pt) != -1)
			error_at_point(&pt, "end of file in conditional");
		break;
	}

	case ALT:
		error_at_point(start_point(),
			       "end of file in variable expansion");
		break;

	case DQ:
		error_at_point(start_point(),
			       "end of file in variable expansion: "
			       "missing closing double quote");
		break;

	case SQ:
		error_at_point(start_point(),
			       "end of file in variable expansion: "
			       "missing closing single quote");
		break;

	case DSIGIL:
		fputc(sigil, yyout);
	case SIGIL:
		fputc(sigil, yyout);
		break;

	case COMMENT:
		error_at_point(start_point(), "end of file in comment");
		break;

	case FALSE:
		error_at_point(start_point(), "end of file in conditional");
		break;

	case VERBATIM:
		error_at_point(start_point(), "end of file in verbatim block");
		break;

	case SETEXP:
		error_at_point(start_point(), 
			       "end of file in %c%cset statement "
			       "(missing closing double quote?)",
			       sigil, sigil);
		break;

	case SETVAL:
		error_at_point(start_point(), 
			       "end of file in %c%cset statement "
			       "(missing closing single quote?)",
			       sigil, sigil);
		break;
	}
	return status;
}
